import click

from hookee.climanager import CliManager

from future.utils import iteritems

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"


@click.group(invoke_without_command=True)
@click.pass_context
@click.option("--port", type=int, help="The port for the local webserver.")
@click.option("--plugins-dir", type=click.Path(exists=True), help="The directory to scan for custom `hookee` plugins.")
@click.option("--plugins", multiple=True, help="A list of `hookee` plugins to use.")
@click.option("--auth-token", help="The `ngrok` auth token use.")
@click.option("--region", type=click.Choice(["us", "eu", "ap", "au", "sa", "jp", "in"]),
              help="The `ngrok` region to use.")
@click.option("--subdomain", help="The `ngrok` subdomain token use.")
@click.option("--auth", help="The `ngrok` auth token use for endpoints.")
@click.option("--request-script", type=click.Path(exists=True),
              help="A Python script whose `run(request)` method will be called by the default `/webhook` after all request plugins have run.")
@click.option("--response-script", type=click.Path(exists=True),
              help="A Python script whose `run(request, response)` method will be called by the default `/webhook` after all response plugins have run.")
@click.option("--response", type=str,
              help="Data to set for the response, will override all body data from plugins and `--response_script`.")
@click.option("--content-type", type=str,
              help="The \"Content-Type\" header to set when response body data is given with `--response` (defaults to \"text/plain\")")
def hookee(ctx, **kwargs):
    """
    If options are given, they override the default values derived from the config file.

    `hookee` documentation can be found at https://hookee.readthedocs.io.
    """
    ctx.ensure_object(dict)
    for key, value in iteritems(kwargs):
        if value:
            ctx.obj[key] = value

    manager = CliManager(ctx, load_plugins=ctx.invoked_subcommand not in ["enable-plugin", "disable-plugin"])
    ctx.obj["cli_manager"] = manager

    if ctx.invoked_subcommand is None:
        manager.start()


@hookee.command()
@click.pass_context
def start(ctx):
    """
    Start `hookee`.
    """
    manager = ctx.obj["cli_manager"]

    manager.start()


@hookee.command()
@click.pass_context
@click.argument("key")
@click.argument("value")
def update_config(ctx, key, value):
    """
    Update the default value for a config.
    """
    manager = ctx.obj["cli_manager"]

    try:
        manager.config_data.set(key, value)

        manager.print_util.print_config_update(
            "The default value for \"{}\" has been updated in the config.".format(key))
    except KeyError:
        ctx.fail("No such key exists in the config: {}".format(key))


@hookee.command()
@click.pass_context
@click.argument("plugin")
def enable_plugin(ctx, plugin):
    """
    Enable the given plugin by default.
    """
    cli_manager = ctx.obj["cli_manager"]

    loaded_plugin = cli_manager.plugin_manager.source.load_plugin(plugin)

    cli_manager.plugin_manager.validate_plugin(loaded_plugin)

    cli_manager.config_data.append("plugins", plugin)

    cli_manager.print_util.print_config_update("Plugin \"{}\" has been enabled.".format(plugin))


@hookee.command()
@click.pass_context
@click.argument("plugin")
def disable_plugin(ctx, plugin):
    """
    Disable the given plugin by default.
    """
    manager = ctx.obj["cli_manager"]

    if plugin in manager.plugin_manager.REQUIRED_PLUGINS:
        ctx.fail("Sorry, you can't disable the plugin {}.".format(plugin))

    manager.config_data.remove("plugins", plugin)

    manager.print_util.print_config_update("Plugin \"{}\" has been disabled.".format(plugin))


@hookee.command()
@click.pass_context
def list_plugins(ctx):
    """
    List all available plugins.
    """
    cli_manager = ctx.obj["cli_manager"]

    plugins = sorted(cli_manager.plugin_manager.source.list_plugins())

    click.secho("\nAvailable Plugins: {}\n".format(plugins),
                fg="green")


@hookee.command()
@click.pass_context
def enabled_plugins(ctx):
    """
    List all enabled plugins.
    """
    cli_manager = ctx.obj["cli_manager"]

    plugins = sorted(cli_manager.config_data.get("plugins"))

    click.secho("\nEnabled Plugins: {}\n".format(plugins), fg="green")


if __name__ == "__main__":
    hookee(obj={})



from hookee import pluginmanager

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.7"

plugin_type = pluginmanager.REQUEST_PLUGIN
print_util = None


def setup(cli_manager):
    global print_util

    print_util = cli_manager.print_util


def run(request):
    if request.files:
        print_util.print_dict("Files", dict(request.files), fg="magenta")

    return request



import inspect
import json
import sys
import xml.dom.minidom

import click

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"


class PrintUtil:
    """
    An object that provides helpful methods for printing to the console through ``click`` CLI.

    :var config: The ``hookee`` configuration.
    :vartype config: Config
    """
    def __init__(self, config):
        self.config = config

    @property
    def console_width(self):
        return self.config.get("console_width")

    def print_config_update(self, msg):
        click.secho("\n--> {}\n".format(msg), fg="green")

    def print_open_header(self, title, delimiter="-", fg="green"):
        """
        Print an opening header to the CLI.

        :param delimiter: The title of the XML blob.
        :type delimiter: str
        :param fg: The color to make the text.
        :type fg: str
        """
        width = int((self.console_width - len(title)) / 2)

        click.echo("")
        click.secho("{}{}{}".format(delimiter * width, title, delimiter * width), fg=fg, bold=True)
        click.echo("")

    def print_close_header(self, delimiter="-", fg="green"):
        """
        Print a closing header to the CLI.

        :param delimiter: The title of the XML blob.
        :type delimiter: str
        :param fg: The color to make the text.
        :type fg: str
        """
        click.secho(delimiter * self.console_width, fg=fg, bold=True)

    def print_dict(self, title, data, fg="green"):
        """
        Print formatted dictionary data to the CLI.

        :param title: The title of the XML blob.
        :type title: str
        :param data: A dictionary.
        :type data: dict
        :param fg: The color to make the text.
        :type fg: str
        """
        click.secho("{}: {}".format(title, json.dumps(data, indent=4)), fg=fg)

    def print_xml(self, title, data, fg="green"):
        """
        Print formatted XML to the CLI.

        :param title: The title of the XML blob.
        :type title: str
        :param data: An XML string.
        :type data: str
        :param fg: The color to make the text.
        :type fg: str
        """
        click.secho("{}: {}".format(title, xml.dom.minidom.parseString(data).toprettyxml()), fg=fg)


def is_python_3():
    """
    Check if running on a Python 3.x interpreter.

    :return: True if Python 3, False otherwise.
    :rtype: bool
    """
    return sys.version_info >= (3, 0)


def get_functions(mod):
    """
    Get a list of functions for the given module.

    :param mod: The module to inspect for functions.
    :type mod: module
    :return: The list of functions.
    :rtype: list[function]
    """
    return [o[0] for o in inspect.getmembers(mod, inspect.isfunction)]


def get_args(func):
    """
    Get a list of args for the given function.

    :param func: The function to inspect for args.
    :type func: function
    :return: The list of args.
    :rtype: list[str]
    """
    if is_python_3():
        return inspect.getfullargspec(func)[0]
    else:
        return inspect.getargspec(func)[0]



import time

import click

from hookee import conf
from hookee.pluginmanager import PluginManager
from hookee.server import Server
from hookee.tunnel import Tunnel
from hookee.util import PrintUtil

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"


class CliManager:
    """
    An object that manages the state of a CLI application. Reads application configuration, loads enabled plugins,
    and manages the long-lived state of the application if a server and tunnel are started.

    :var ctx: The ``click`` CLI context.
    :vartype ctx: click.Context
    :var config: The ``hookee`` configuration.
    :vartype config: Config
    :var plugin_manager: Reference to the Plugin Manager.
    :vartype plugin_manager: PluginManager
    :var print_util: Reference to the PrintUtil.
    :vartype print_util: PrintUtil
    :var tunnel: Reference to the Tunnel.
    :vartype tunnel: Tunnel
    :var server: Reference to the Server.
    :vartype server: Server
    :var alive: ``True`` when this object is managing an active tunnel and server, ``False`` otherwise.
    :vartype alive: bool
    """

    def __init__(self, ctx, load_plugins=True):
        self.ctx = ctx

        self.config = conf.Config(self.ctx)
        self.plugin_manager = PluginManager(self)
        self.print_util = PrintUtil(self.config)

        if load_plugins:
            self.plugin_manager.load_plugins()

        self.tunnel = Tunnel(self)
        self.server = Server(self)

        self.alive = False

        self.print_hookee_banner()

    def start(self):
        """
        If one is not already running, start a managed server and tunnel and block until an interrupt
        is received (or ``alive`` is set to False).
        """
        if not self.alive:
            try:
                self._init_server_and_tunnel()

                while self.alive:
                    time.sleep(1)
            except KeyboardInterrupt:
                pass

            self.stop()

    def stop(self):
        """
        If running, shutdown server and tunnel.
        """
        if self.alive:
            self.server.stop()
            if self.tunnel._thread:
                self.tunnel._thread.alive = False

            # Wait for the other threads to teardown
            while self.server._thread and self.tunnel._thread:
                time.sleep(1)

            self.alive = False

    def print_hookee_banner(self):
        self.print_util.print_open_header("", "=")
        click.secho("""                .__                   __                  
                |  |__   ____   ____ |  | __ ____   ____  
                |  |  \ /  _ \ /  _ \|  |/ // __ \_/ __ \ 
                |   Y  (  <_> |  <_> )    <\  ___/\  ___/ 
                |___|  /\____/ \____/|__|_ \\___  >\___  >
                     \/                   \/    \/     \/ 
                                                   v{}""".format(__version__), fg="green", bold=True)
        self.print_util.print_close_header("=")

    def print_ready(self):
        self.print_util.print_open_header("Registered Endpoints")

        rules = list(filter(lambda r: r.rule not in ["/shutdown", "/static/<path:filename>", "/status"],
                            self.server.app.url_map.iter_rules()))
        for rule in rules:
            click.secho("* {}{}".format(self.tunnel.public_url, rule.rule))
            click.secho("  Methods: {}".format(sorted(list(rule.methods))))
            click.echo("")

        self.print_util.print_close_header()

        click.echo("")
        click.secho("--> Ready, send a request to a registered endpoint ...", fg="green", bold=True)
        click.echo("")

    def _init_server_and_tunnel(self):
        self.alive = True
        self.server.start()
        self.tunnel.start()

        self.print_ready()



import os

import confuse

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"

template = {
    "auth_token": confuse.String(default=None),
    "region": confuse.Choice(["us", "eu", "ap", "au", "sa", "jp", "in"], default=None),
    "subdomain": confuse.String(default=None),
    "auth": confuse.String(default=None),
    "port": int,
    "plugins_dir": confuse.Filename(),
    "plugins": list,
    "console_width": confuse.Integer(default=80),
    "request_script": confuse.Filename(default=None),
    "response_script": confuse.Filename(default=None),
    "response": confuse.String(default=None),
    "content_type": confuse.String(default=None)
}


class Config:
    """
    An object with accessor methods containing ``hookee``'s configuration. Default configuration can be
    overridden by creating a custom ``config.yaml`` in ``~/.config/hookee`` (when setting config
    values from the command lind, this is where values are updated) which in turn can be overridden by
    passing args to the CLI.

    :var ctx: The ``click`` CLI context.
    :vartype ctx: click.Context
    :var config_obj: The templated config object.
    :vartype config_obj: confuse.core.Configuration
    :var config_dir: The directory of the config being used.
    :vartype config_dir: str
    :var config_filename: The full path to the config file being used.
    :vartype config_filename: str
    :var config_data: The parsed and validated config data. Use :func:`get`, :func:`set`, and other accessors
        to interact with the data.
    :vartype config_data: confuse.templates.AttrDict
    """

    def __init__(self, ctx):
        self.ctx = ctx

        try:
            config = confuse.Configuration("hookee", __name__)
            config.set_args(ctx.obj)

            self.config_obj = config
            self.config_dir = self.config_obj.config_dir()
            self.config_path = os.path.join(self.config_dir, confuse.CONFIG_FILENAME)

            self.config_data = config.get(template)

            plugins_dir = os.path.expanduser(self.config_data["plugins_dir"])
            if not os.path.exists(plugins_dir):
                os.makedirs(plugins_dir)
        except confuse.NotFoundError as e:
            ctx.fail("The config file is invalid: {}.".format(str(e)))
        except (confuse.ConfigReadError, ValueError):
            ctx.fail("The config file is not valid YAML.")

    def get(self, key):
        """
        Get the config value for the given key.

        :param key: The key.
        :type key: str
        :return: The config value.
        :rtype: object
        """
        return self.config_data[key]

    def set(self, key, value):
        """
        Update the config value for the given key, persisting to user's ``config.yaml``.

        :param key: The key.
        :type key: str
        :param value: The value to set.
        :type key: object
        """
        if value != self.config_data[key]:
            self.config_data[key] = value

            self.config_obj[key] = value
            with open(self.config_path, "w") as f:
                f.write(self.config_obj.dump())

    def append(self, key, value):
        """
        Update the config value by appending to the list for the given key, persisting to
        user's ``config.yaml``.

        :param key: The key.
        :type key: str
        :param value: The value to append.
        :type value: object
        """
        list_item = list(self.config_data[key])

        if value not in list_item:
            list_item.append(value)
            self.config_data[key] = list_item

            self.config_obj[key] = list_item
            with open(self.config_path, "w") as f:
                f.write(self.config_obj.dump())

    def remove(self, key, value):
        """
        Update the config value by removing the given value from the list for the given key, persisting to
        user's ``config.yaml``.

        :param key: The key.
        :type key: str
        :param value: The value to remove.
        :type value: object
        """
        list_item = list(self.config_data[key])

        if value in list_item:
            list_item.remove(value)
            self.config_data[key] = list_item

            self.config_obj[key] = list_item
            with open(self.config_path, "w") as f:
                f.write(self.config_obj.dump())



from flask import jsonify

from hookee import pluginmanager

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"

plugin_type = pluginmanager.RESPONSE_PLUGIN
print_util = None


def setup(cli_manager):
    global print_util

    print_util = cli_manager.print_util


def run(request, response):
    if not response:
        response = jsonify({"Response": "Ok"})

    return response



import threading
import time

import click

from pyngrok import ngrok
from pyngrok.conf import PyngrokConfig

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"


class Tunnel:
    """
    An object that manages a non-blocking ``pyngrok`` tunnel thread.

    :var cli_manager: Reference to the CLI Manager.
    :vartype cli_manager: CliManager
    :var plugin_manager: Reference to the Plugin Manager.
    :vartype plugin_manager: PluginManager
    :var print_util: Reference to the PrintUtil.
    :vartype print_util: PrintUtil
    :var port: The server's port.
    :vartype port: int
    :var pyngrok_config: The ``pyngrok`` config.
    :vartype pyngrok_config: pyngrok.conf.PyngrokConfig
    :var public_url: The public URL of the tunnel.
    :vartype public_url: str
    :var ngrok_process: The ``ngrok`` process.
    :vartype ngrok_process: pyngrok.process.NgrokProcess
    """

    def __init__(self, cli_manager):
        self.cli_manager = cli_manager
        self.config = cli_manager.config
        self.plugin_manager = cli_manager.plugin_manager
        self.print_util = cli_manager.print_util
        self.port = self.config.get("port")

        self.pyngrok_config = PyngrokConfig(auth_token=self.config.get("auth_token"),
                                            region=self.config.get("region"))

        self.public_url = None
        self.ngrok_process = None
        self._thread = None

    def _loop(self):
        self._start_tunnel()

        thread = threading.current_thread()
        thread.alive = True
        while thread.alive:
            time.sleep(1)

        thread.alive = False

        self.stop()

    def start(self):
        """
        If one is not already running, start a tunnel in a new thread.
        """
        if self._thread is None:
            self.print_util.print_open_header("Opening Tunnel")

            self._thread = threading.Thread(target=self._loop)
            self._thread.daemon = True
            self._thread.start()

            while self.public_url is None:
                time.sleep(1)

            self.print_close_header()

    def _start_tunnel(self):
        options = {}
        subdomain = self.config.get("subdomain")
        auth = self.config.get("auth")
        if subdomain:
            options["subdomain"] = subdomain
        if auth:
            options["auth"] = auth

        self.public_url = ngrok.connect(self.port,
                                        pyngrok_config=self.pyngrok_config,
                                        options=options).replace("http", "https")
        self.ngrok_process = ngrok.get_ngrok_process()

    def stop(self):
        """
        If running, kill the tunnel and cleanup its thread.
        """
        if self._thread:
            ngrok.kill()

            self.public_url = None
            self.ngrok_process = None
            self._thread = None

    def print_close_header(self):
        click.echo(
            "* Tunnel: {} -> http://127.0.0.1:{}".format(self.public_url.replace("http://", "https://"), self.port))
        click.echo("")
        self.print_util.print_close_header()



from datetime import datetime

import click

from hookee import pluginmanager

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.7"

plugin_type = pluginmanager.REQUEST_PLUGIN
print_util = None


def setup(cli_manager):
    global print_util

    print_util = cli_manager.print_util


def run(request):
    now = datetime.now()

    click.secho(
        "[{}] \"{} {} {}\"".format(now.isoformat(), request.method, request.base_url,
                                   request.environ["SERVER_PROTOCOL"]),
        fg="magenta", bold=True)
    click.echo("")

    return request



import os

from hookee import util

from pluginbase import PluginBase

if util.is_python_3():
    import importlib.util
else:
    import imp

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"

BLUEPRINT_PLUGIN = "blueprint"
REQUEST_PLUGIN = "request"
RESPONSE_PLUGIN = "response"

VALID_PLUGIN_TYPES = [BLUEPRINT_PLUGIN, REQUEST_PLUGIN, RESPONSE_PLUGIN]
REQUIRED_PLUGINS = ["blueprint_default"]


class PluginManager:
    """
    An object that loads, validates, and manages available and enabled plugins.

    :var cli_manager: Reference to the CLI Manager.
    :vartype cli_manager: CliManager
    :var ctx: The ``click`` CLI context.
    :vartype ctx: click.Context
    :var config: The ``hookee`` configuration.
    :vartype config: Config
    :var source: The ``hookee`` configuration.
    :vartype source: pluginbase.PluginSource
    :var request_script: A request plugin loaded from the script at ``--request_script``, run last.
    :vartype request_script: module
    :var response_script: A response plugin loaded from the script at ``--response_script``, run last.
    :vartype response_script: module
    :var response_body: The response body loaded from ``--response``, overrides any body data from response plugins.
    :vartype response_body: module
    :var builtin_plugins_dir: The directory where built-in plugins reside.
    :vartype builtin_plugins_dir: str
    :var loaded_plugins: A list of plugins that have been validated and imported.
    :vartype loaded_plugins: list[module]
    """

    def __init__(self, cli_manager):
        self.cli_manager = cli_manager
        self.ctx = cli_manager.ctx
        self.config = cli_manager.config

        self.source = None
        self.request_script = None
        self.response_script = None
        self.response_body = None
        self.response_content_type = None

        self.builtin_plugins_dir = os.path.normpath(os.path.join(os.path.abspath(os.path.dirname(__file__)), "plugins"))

        self.loaded_plugins = []

        self.source_plugins()

    def validate_plugin(self, plugin):
        """
        Validate a given module to see if it is a valid ``hookee`` plugin. If validation fails, an exception
        will be thrown.

        :param plugin: The module to validate as a valid plugin.
        :type plugin: module
        :return: ``True`` if the validated plugin has a ``setup()`` method, ``False`` otherwise.
        :type: bool
        """
        try:
            functions_list = util.get_functions(plugin)
            attributes = dir(plugin)

            if "plugin_type" not in attributes:
                self.ctx.fail("Plugin \"{}\" does not conform to the plugin spec.".format(self.get_plugin_name(plugin)))
            elif plugin.plugin_type not in VALID_PLUGIN_TYPES:
                self.ctx.fail("Plugin \"{}\" must specify a valid `plugin_type`.".format(self.get_plugin_name(plugin)))
            elif plugin.plugin_type == REQUEST_PLUGIN:
                if "run" not in functions_list:
                    self.ctx.fail("Plugin \"{}\" must implement run(request).".format(self.get_plugin_name(plugin)))
                elif len(util.get_args(plugin.run)[0]) < 1:
                    self.ctx.fail(
                        "Plugin \"{}\" does not conform to the plugin spec, `run(request)` must be defined.".format(
                            self.get_plugin_name(plugin)))
            elif plugin.plugin_type == RESPONSE_PLUGIN:
                if "run" not in functions_list:
                    self.ctx.fail(
                        "Plugin \"{}\" must implement run(request, response).".format(self.get_plugin_name(plugin)))
                elif len(util.get_args(plugin.run)[0]) < 2:
                    self.ctx.fail(
                        "Plugin \"{}\" does not conform to the plugin spec, `run(request, response)` must be defined.".format(
                            self.get_plugin_name(plugin)))
            elif plugin.plugin_type == BLUEPRINT_PLUGIN and "blueprint" not in attributes:
                self.ctx.fail(
                    "Plugin \"{}\" must define `blueprint = Blueprint(\"plugin_name\", __name__)`.".format(
                        self.get_plugin_name(plugin)))

            return "setup" in functions_list
        except ModuleNotFoundError:
            self.ctx.fail("Plugin \"{}\" could not be found.".format(self.get_plugin_name(plugin)))

    def source_plugins(self):
        """
        Source all paths in the plugin base to prepare for loading and validating plugins.
        """
        plugins_dir = self.config.get("plugins_dir")

        plugin_base = PluginBase(package="hookee.plugins",
                                 searchpath=[self.builtin_plugins_dir])
        self.source = plugin_base.make_plugin_source(searchpath=[plugins_dir])

    def load_plugins(self):
        """
        Load and validate all built-in plugins and custom plugins from sources in the plugin base.
        """
        enabled_plugins = self.config.get("plugins")

        for plugin_name in REQUIRED_PLUGINS:
            if plugin_name not in enabled_plugins:
                self.ctx.fail(
                    "Sorry, the plugin {} is required. Run `hookee enable-plugin {}` before continuing.".format(
                        plugin_name, plugin_name))

        self.source_plugins()

        self.loaded_plugins = []
        for plugin_name in enabled_plugins:
            plugin = self.source.load_plugin(plugin_name)
            if self.validate_plugin(plugin):
                plugin.setup(self.cli_manager)
            self.loaded_plugins.append(plugin)

        self.request_script = self.import_from_file(self.config.get("request_script"))
        self.response_script = self.import_from_file(self.config.get("response_script"))
        self.response_body = self.config.get("response")
        self.response_content_type = self.config.get("content_type")

        if self.response_content_type and not self.response_body:
            self.ctx.fail("If `--content_type` is given, `--response` must also be given.")

    def get_plugin_name(self, plugin):
        """
        Get the name of the plugin from the module.

        :param plugin: The plugin.
        :type plugin: module
        :return: The name of the plugin.
        :rtype: str
        """
        return os.path.basename(plugin.__file__).strip(".py")

    def get_plugins_by_type(self, plugin_type):
        """
        Filter loaded plugins by the given plugin type.

        :param plugin_type: The plugin type for filtering.
        :type plugin_type: str
        :return: The filtered list of plugins.
        :rtype: list[module]
        """
        return filter(lambda p: p.plugin_type == plugin_type, self.loaded_plugins)

    def import_from_file(self, path):
        """
        Import a Python script at the given path as a module that is executable.

        :param path: The path to the script to import.
        :type path: str
        :return: The imported script as a module.
        :rtype: module
        """
        if path:
            module_name = os.path.basename(path).strip(".py")

            if util.is_python_3():
                spec = importlib.util.spec_from_file_location(module_name, path)
                plugin = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(plugin)
            else:
                plugin = imp.load_source(module_name, path)

            if self.validate_plugin(plugin):
                plugin.setup(self.cli_manager)

            return plugin

    def run_request_plugins(self, request):
        """
        Run all enabled request plugins.

        :param request: The request object being processed.
        :type request: flask.Request
        :return: The processed request.
        :rtype: flask.Request
        """
        for plugin in self.get_plugins_by_type(REQUEST_PLUGIN):
            request = plugin.run(request)
        if self.request_script:
            self.request_script.run(request)

        return request

    def run_response_plugins(self, request=None, response=None):
        """
        Run all enabled response plugins, running the ``response_info`` plugin (if enabled) last.

        :param request: The request object being processed.
        :type request: flask.Request, optional
        :param response: The response object being processed.
        :type response: flask.Response, optional
        :return: The processed response.
        :rtype: flask.Response
        """
        response_info_plugin = None
        for plugin in self.get_plugins_by_type(RESPONSE_PLUGIN):
            if self.get_plugin_name(plugin) == "response_info":
                response_info_plugin = plugin
            else:
                response = plugin.run(request, response)
        if self.response_script:
            response = self.response_script.run(request, response)
        if self.response_body:
            response.data = self.response_body
            response.headers[
                "Content-Type"] = self.response_content_type if self.response_content_type else "text/plain"
        if response_info_plugin:
            response = response_info_plugin.run(request, response)

        return response



import click

from hookee import pluginmanager

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.7"

plugin_type = pluginmanager.REQUEST_PLUGIN
print_util = None


def setup(cli_manager):
    global print_util

    print_util = cli_manager.print_util


def run(request):
    if request.is_json:
        click.secho("Body Type: JSON", fg="magenta")
        print_util.print_dict("Body", dict(request.json), fg="magenta")
    elif request.form and not request.data:
        click.secho("Body Type: FORM")
        print_util.print_dict("Body", dict(request.form), fg="magenta")
    elif request.data:
        click.secho("Body: {}".format(request.data.decode("utf-8")), fg="magenta")

    return request



import click

from hookee import pluginmanager

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"

plugin_type = pluginmanager.RESPONSE_PLUGIN
print_util = None


def setup(cli_manager):
    global print_util

    print_util = cli_manager.print_util


def run(request, response):
    click.secho("Status Code: {}".format(response.status_code), fg="magenta")
    if response.headers:
        print_util.print_dict("Headers", dict(response.headers), fg="magenta")
    if response.data:
        if response.is_json:
            print_util.print_dict("Body", response.get_json(), fg="magenta")
        else:
            click.secho("Body: {}".format(response.data.decode("utf-8")), fg="magenta")

    return response



import logging
import threading
import time

import click

from flask import Flask

from future.standard_library import install_aliases

from hookee import pluginmanager

install_aliases()

from urllib.request import urlopen, Request
from urllib.error import URLError

try:
    from http import HTTPStatus as StatusCodes
except ImportError:  # pragma: no cover
    try:
        from http import client as StatusCodes
    except ImportError:
        import httplib as StatusCodes

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"

werkzeug_logger = logging.getLogger('werkzeug')
werkzeug_logger.setLevel(logging.ERROR)


class Server:
    """
    An object that manages a non-blocking Flask server thread.

    :var cli_manager: Reference to the CLI Manager.
    :vartype cli_manager: CliManager
    :var plugin_manager: Reference to the Plugin Manager.
    :vartype plugin_manager: PluginManager
    :var print_util: Reference to the PrintUtil.
    :vartype print_util: PrintUtil
    :var port: The server's port.
    :vartype port: int
    :var app: The Flask app.
    :vartype app: flask.Flask
    """

    def __init__(self, cli_manager):
        self.cli_manager = cli_manager
        self.plugin_manager = cli_manager.plugin_manager
        self.print_util = cli_manager.print_util
        self.port = self.cli_manager.config.get("port")

        self.app = self.create_app()

        self._thread = None

    def create_app(self):
        """
        Create a Flask app and register all Blueprints found in enabled plugins.

        :return: The Flask app.
        :rtype: flask.Flask
        """
        app = Flask(__name__)

        app.config.from_mapping(
            ENV="development"
        )

        for plugin in self.plugin_manager.get_plugins_by_type(pluginmanager.BLUEPRINT_PLUGIN):
            app.register_blueprint(plugin.blueprint)

        return app

    def _loop(self):
        thread = threading.current_thread()
        thread.alive = True

        # This will block until stop() is invoked to shutdown the Werkzeug server
        self.app.run(host="127.0.0.1", port=self.port, debug=True, use_reloader=False)

        thread.alive = False

    def start(self):
        """
        If one is not already running, start a server in a new thread.
        """
        if self._thread is None:
            self.print_util.print_open_header("Starting Server")

            self._thread = threading.Thread(target=self._loop)
            self._thread.start()

            while self._server_status() != StatusCodes.OK:
                time.sleep(1)

            self.print_close_header()

    def stop(self):
        """
        If running, kill the server and cleanup its thread.
        """
        if self._thread:
            req = Request("http://127.0.0.1:{}/shutdown".format(self.port), method="POST")
            urlopen(req)

            self._thread = None

    def _server_status(self):
        """
        Get the status code of the server's ``/status`` endpoint.

        :return: The status code.
        :rtype: http.HTTPStatus
        """
        try:
            return urlopen("http://127.0.0.1:{}/status".format(self.port)).getcode()
        except URLError:
            return StatusCodes.INTERNAL_SERVER_ERROR

    def print_close_header(self):
        click.echo(" * Port: {}".format(self.port))
        click.echo(" * Blueprints: registered")
        click.echo("")
        self.print_util.print_close_header()



import click

from flask import Blueprint, request

from hookee import pluginmanager

__author__ = "Alex Laird"
__copyright__ = "Copyright 2020, Alex Laird"
__version__ = "0.0.8"

blueprint = Blueprint("default", __name__)

plugin_type = pluginmanager.BLUEPRINT_PLUGIN
plugin_manager = None
print_util = None


def setup(cli_manager):
    global plugin_manager, print_util

    plugin_manager = cli_manager.plugin_manager
    print_util = cli_manager.print_util


@blueprint.route("/webhook",
                 methods=["GET", "HEAD", "POST", "PUT", "DELETE", "PATCH", "OPTIONS", "TRACE", "CONNECT"])
def webhook():
    print_util.print_close_header(delimiter="=", fg="magenta")

    print_util.print_open_header("Request", delimiter="-", fg="magenta")

    plugin_manager.run_request_plugins(request)

    print_util.print_open_header("Response", fg="magenta")

    response = plugin_manager.run_response_plugins(request)

    click.echo("")
    print_util.print_close_header("=", fg="magenta")
    click.echo("")

    return response


@blueprint.route("/status")
def status():
    return "", 200


@blueprint.route("/shutdown", methods=["POST"])
def shutdown():
    request.environ.get("werkzeug.server.shutdown")()

    return "", 204
